@@ -31,20 +31,6 @@
 
 #include <config.h>
 
-/* For >= 500 for pread/pwrite; >= 700 for utimensat */
-#define _XOPEN_SOURCE 700
-
-/* For flock() on FreeBSD. It otherwise gets hidden by _XOPEN_SOURCE  */
-#define __BSD_VISIBLE 1
-
-/* For stat() nanosecond precision and lutimes() */
-#define _BSD_SOURCE
-/* The new non-deprecated version of _BSD_SOURCE */
-#define _DEFAULT_SOURCE
-
-/* Fix MacOS realpath() broken around Catalina (#83) */
-#define _DARWIN_BETTER_REALPATH
-
 #include <stdlib.h>
 #include <stddef.h>
 #include <stdio.h>
@@ -57,6 +43,9 @@
 #ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
+#ifdef HAVE_SYS_FILE_H
+#include <sys/file.h>
+#endif
 #include <sys/time.h>
 #include <sys/statvfs.h>
 #include <sys/file.h>
@@ -75,6 +64,12 @@
 #include <sys/xattr.h>
 #endif
 
+#ifdef HAVE_FUSE_3
+#ifndef __NR_renameat2
+#include <libgen.h> // For dirname(), basename()
+#endif
+#endif
+
 #ifdef __linux__
 #include <sys/mman.h>
 #include <fcntl.h>
@@ -88,6 +83,7 @@
 #include <fuse.h>
 #include <fuse_opt.h>
 
+#include "arena.h"
 #include "debug.h"
 #include "misc.h"
 #include "permchain.h"
@@ -103,6 +99,13 @@
 #define A_PREFIX   "com"
 #define A_KAUTH_FILESEC_XATTR A_PREFIX ".apple.system.Security"
 #define XATTR_APPLE_PREFIX   "com.apple."
+
+// Yes, Apple asks us to copy/paste these -.-
+#define   LOCK_SH   1    /* shared lock */
+#define   LOCK_EX   2    /* exclusive lock */
+#define   LOCK_NB   4    /* don't block when locking */
+#define   LOCK_UN   8    /* unlock */
+int flock(int fd, int operation);
 #endif
 
 /* We pessimistically assume signed uid_t and gid_t in our overflow checks,
@@ -207,6 +210,8 @@ static struct Settings {
 
 } settings;
 
+static bool bindfs_init_failed = false;
+
 
 
 /* PROTOTYPES */
@@ -216,6 +221,13 @@ static int is_mirroring_enabled();
 /* Checks whether the uid is to be the mirrored owner of all files. */
 static int is_mirrored_user(uid_t uid);
 
+#ifdef HAVE_FUSE_3
+#ifndef __NR_renameat2
+/* Returns the path to a temporary file inside the same directory as src_path */
+char *tmpnam_inpath(char *src_path);
+#endif
+#endif
+
 /* Processes the virtual path to a real path. Always free() the result. */
 static char *process_path(const char *path, bool resolve_symlinks);
 
@@ -239,27 +251,62 @@ static size_t round_up_buffer_size_for_direct_io(size_t size);
 #endif
 
 /* FUSE callbacks */
+#ifdef HAVE_FUSE_3
+static void *bindfs_init(struct fuse_conn_info *conn, struct fuse_config *cfg);
+#else
 static void *bindfs_init();
+#endif
 static void bindfs_destroy(void *private_data);
+#ifdef HAVE_FUSE_3
+static int bindfs_getattr(const char *path, struct stat *stbuf,
+                           struct fuse_file_info *fi);
+#else
 static int bindfs_getattr(const char *path, struct stat *stbuf);
 static int bindfs_fgetattr(const char *path, struct stat *stbuf,
                            struct fuse_file_info *fi);
+#endif
 static int bindfs_readlink(const char *path, char *buf, size_t size);
+#ifdef HAVE_FUSE_3
+static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
+                          off_t offset, struct fuse_file_info *fi, enum fuse_readdir_flags flags);
+#else
 static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                           off_t offset, struct fuse_file_info *fi);
+#endif
 static int bindfs_mknod(const char *path, mode_t mode, dev_t rdev);
 static int bindfs_mkdir(const char *path, mode_t mode);
 static int bindfs_unlink(const char *path);
 static int bindfs_rmdir(const char *path);
 static int bindfs_symlink(const char *from, const char *to);
+#ifdef HAVE_FUSE_3
+static int bindfs_rename(const char *from, const char *to, unsigned int flags);
+#else
 static int bindfs_rename(const char *from, const char *to);
+#endif
 static int bindfs_link(const char *from, const char *to);
+#ifdef HAVE_FUSE_3
+static int bindfs_chmod(const char *path, mode_t mode, struct fuse_file_info *fi);
+#else
 static int bindfs_chmod(const char *path, mode_t mode);
+#endif
+#ifdef HAVE_FUSE_3
+static int bindfs_chown(const char *path, uid_t uid, gid_t gid, struct fuse_file_info *fi);
+#else
 static int bindfs_chown(const char *path, uid_t uid, gid_t gid);
+#endif
+#ifdef HAVE_FUSE_3
+static int bindfs_truncate(const char *path, off_t size,
+                            struct fuse_file_info *fi);
+#else
 static int bindfs_truncate(const char *path, off_t size);
 static int bindfs_ftruncate(const char *path, off_t size,
                             struct fuse_file_info *fi);
+#endif
+#ifdef HAVE_FUSE_3
+static int bindfs_utimens(const char *path, const struct timespec tv[2], struct fuse_file_info *fi);
+#else
 static int bindfs_utimens(const char *path, const struct timespec tv[2]);
+#endif
 static int bindfs_create(const char *path, mode_t mode, struct fuse_file_info *fi);
 static int bindfs_open(const char *path, struct fuse_file_info *fi);
 static int bindfs_read(const char *path, char *buf, size_t size, off_t offset,
@@ -269,9 +316,15 @@ static int bindfs_write(const char *path, const char *buf, size_t size,
 static int bindfs_lock(const char *path, struct fuse_file_info *fi, int cmd,
                        struct flock *lock);
 static int bindfs_flock(const char *path, struct fuse_file_info *fi, int op);
+#ifdef HAVE_FUSE_3
 static int bindfs_ioctl(const char *path, int cmd, void *arg,
                         struct fuse_file_info *fi, unsigned int flags,
                         void *data);
+#else
+static int bindfs_ioctl(const char *path, int cmd, void *arg,
+                        struct fuse_file_info *fi, unsigned int flags,
+                        void *data);
+#endif
 static int bindfs_statfs(const char *path, struct statvfs *stbuf);
 static int bindfs_release(const char *path, struct fuse_file_info *fi);
 static int bindfs_fsync(const char *path, int isdatasync,
@@ -293,6 +346,16 @@ static void signal_handler(int sig);
 
 static void atexit_func();
 
+/* 
+Ignore some options (starting with -o) 
+like "mount.fuse" would do in "libfuse/util/mount.fuse.c" 
+
+Why? Because some of those options like "nofail"
+are special ones interpreted by systemd in /etc/fstab
+*/
+struct fuse_args filter_special_opts(struct fuse_args *args);
+static bool keep_option(const char* opt);
+
 static int is_mirroring_enabled()
 {
     return settings.num_mirrored_users + settings.num_mirrored_members > 0;
@@ -314,6 +377,65 @@ static int is_mirrored_user(uid_t uid)
     return 0;
 }
 
+#ifdef HAVE_FUSE_3
+#ifndef __NR_renameat2
+char *tmpnam_inpath(char *src_path) 
+{
+    
+    char *res, tmp[L_tmpnam], *path, *fname, *tfname, *src_path_cpy, *src_path_cpy_1, *tmp_cpy;
+    int i = 0;
+    
+    src_path_cpy = (char *) calloc((strlen(src_path)), sizeof(char)); 
+    strcpy(src_path_cpy, src_path);
+    
+    src_path_cpy_1 = (char *) calloc((strlen(src_path)), sizeof(char)); 
+    strcpy(src_path_cpy_1, src_path);
+    
+    path = dirname(src_path_cpy);
+    fname = basename(src_path_cpy_1);
+    
+    tmpnam(tmp);
+    tmp_cpy = (char *) calloc((strlen(tmp)), sizeof(char));
+    strcpy(tmp_cpy, tmp);
+    
+    tfname = basename(tmp_cpy);
+    
+    res = (char *) calloc( (strlen(path) + 2 + strlen(tfname) + 1 + strlen(fname) + 1), sizeof(char));
+    
+    strcat(res, path);
+    strcat(res, "/.");
+    strcat(res, tfname);
+    strcat(res, "~");
+    strcat(res, fname);
+    
+    while (access( res, F_OK ) == 0 && i < 10) {
+        
+        tmpnam(tmp);
+        tmp_cpy = (char *) calloc((strlen(tmp)), sizeof(char));
+        strcpy(tmp_cpy, tmp);
+    
+        tfname = basename(tmp_cpy);
+    
+        res = (char *) calloc( (strlen(path) + 2 + strlen(tfname) + 1 + strlen(fname) + 1), sizeof(char));
+    
+        strcat(res, path);
+        strcat(res, "/.");
+        strcat(res, tfname);
+        strcat(res, "~");
+        strcat(res, fname);
+
+        i++;
+    }
+    
+    if (access( res, F_OK ) == 0) {
+        return NULL;
+    }
+    
+    return res;
+    
+}
+#endif
+#endif
 
 static char *process_path(const char *path, bool resolve_symlinks)
 {
@@ -618,8 +740,24 @@ static size_t round_up_buffer_size_for_direct_io(size_t size)
 }
 #endif
 
+#ifdef HAVE_FUSE_3
+static void *bindfs_init(struct fuse_conn_info *conn, struct fuse_config *cfg)
+#else
 static void *bindfs_init()
+#endif
 {
+    #ifdef HAVE_FUSE_3
+    (void) conn;
+    cfg->use_ino = 1;
+
+    // Disable caches so changes in base FS are visible immediately.
+    // Especially the attribute cache must be disabled when different users
+    // might see different file attributes, such as when mirroring users.
+    cfg->entry_timeout = 0;
+    cfg->attr_timeout = 0;
+    cfg->negative_timeout = 0;
+    #endif
+
     assert(settings.permchain != NULL);
     assert(settings.mntsrc_fd > 0);
 
@@ -632,6 +770,7 @@ static void *bindfs_init()
             settings.mntsrc,
             strerror(errno)
             );
+        bindfs_init_failed = true;
         fuse_exit(fuse_get_context()->fuse);
     }
 
@@ -642,7 +781,12 @@ static void bindfs_destroy(void *private_data)
 {
 }
 
+#ifdef HAVE_FUSE_3
+static int bindfs_getattr(const char *path, struct stat *stbuf,
+                           struct fuse_file_info *fi)
+#else
 static int bindfs_getattr(const char *path, struct stat *stbuf)
+#endif
 {
     int res;
     char *real_path;
@@ -661,6 +805,7 @@ static int bindfs_getattr(const char *path, struct stat *stbuf)
     return res;
 }
 
+#ifndef HAVE_FUSE_3
 static int bindfs_fgetattr(const char *path, struct stat *stbuf,
                            struct fuse_file_info *fi)
 {
@@ -679,6 +824,7 @@ static int bindfs_fgetattr(const char *path, struct stat *stbuf,
     free(real_path);
     return res;
 }
+#endif
 
 static int bindfs_readlink(const char *path, char *buf, size_t size)
 {
@@ -702,8 +848,13 @@ static int bindfs_readlink(const char *path, char *buf, size_t size)
     return 0;
 }
 
+#ifdef HAVE_FUSE_3
+static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
+                          off_t offset, struct fuse_file_info *fi, enum fuse_readdir_flags flags)
+#else
 static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                           off_t offset, struct fuse_file_info *fi)
+#endif
 {
     char *real_path = process_path(path, true);
     if (real_path == NULL) {
@@ -775,7 +926,13 @@ static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
         // consider it an error if it does. It is undocumented whether it sets
         // errno in that case, so we zero it first and set it ourself if it
         // doesn't.
+        #ifdef HAVE_FUSE_3
+        // TODO: FUSE_FILL_DIR_PLUS doesn't work with offset==0: https://github.com/libfuse/libfuse/issues/583
+        //       Work around this by implementing the offset!=1 mode. Be careful - it's quite error-prone!
+        if (filler(buf, de->d_name, &st, 0, FUSE_FILL_DIR_PLUS) != 0) {
+        #else
         if (filler(buf, de->d_name, &st, 0) != 0) {
+        #endif
             result = errno != 0 ? -errno : -EIO;
             break;
         }
@@ -879,7 +1036,11 @@ static int bindfs_symlink(const char *from, const char *to)
     return res;
 }
 
+#ifdef HAVE_FUSE_3
+static int bindfs_rename(const char *from, const char *to, unsigned int flags)
+#else
 static int bindfs_rename(const char *from, const char *to)
+#endif
 {
     int res;
     char *real_from, *real_to;
@@ -897,7 +1058,25 @@ static int bindfs_rename(const char *from, const char *to)
         return -errno;
     }
 
+#ifdef HAVE_FUSE_3
+
+    if (flags == 0) {
+        res = rename(real_from, real_to);
+    } else {
+#ifdef __NR_renameat2
+        res = syscall(__NR_renameat2, AT_FDCWD, real_from, AT_FDCWD, real_to, flags);
+#else // __NR_renameat2
+        res = -1;
+        errno = EINVAL;
+#endif // __NR_renameat2
+    }
+
+#else  // HAVE_FUSE_3
+
     res = rename(real_from, real_to);
+
+#endif // HAVE_FUSE_3
+
     free(real_from);
     free(real_to);
     if (res == -1)
@@ -930,7 +1109,11 @@ static int bindfs_link(const char *from, const char *to)
     return 0;
 }
 
+#ifdef HAVE_FUSE_3
+static int bindfs_chmod(const char *path, mode_t mode, struct fuse_file_info *fi)
+#else
 static int bindfs_chmod(const char *path, mode_t mode)
+#endif
 {
     int file_execute_only = 0;
     struct stat st;
@@ -993,7 +1176,11 @@ static int bindfs_chmod(const char *path, mode_t mode)
     }
 }
 
+#ifdef HAVE_FUSE_3
+static int bindfs_chown(const char *path, uid_t uid, gid_t gid, struct fuse_file_info *fi)
+#else
 static int bindfs_chown(const char *path, uid_t uid, gid_t gid)
+#endif
 {
     int res;
     char *real_path;
@@ -1044,7 +1231,12 @@ static int bindfs_chown(const char *path, uid_t uid, gid_t gid)
     return 0;
 }
 
+#ifdef HAVE_FUSE_3
+static int bindfs_truncate(const char *path, off_t size,
+                            struct fuse_file_info *fi)
+#else
 static int bindfs_truncate(const char *path, off_t size)
+#endif
 {
     int res;
     char *real_path;
@@ -1061,6 +1253,7 @@ static int bindfs_truncate(const char *path, off_t size)
     return 0;
 }
 
+#ifndef HAVE_FUSE_3
 static int bindfs_ftruncate(const char *path, off_t size,
                             struct fuse_file_info *fi)
 {
@@ -1073,8 +1266,13 @@ static int bindfs_ftruncate(const char *path, off_t size,
 
     return 0;
 }
+#endif
 
+#ifdef HAVE_FUSE_3
+static int bindfs_utimens(const char *path, const struct timespec ts[2], struct fuse_file_info *fi)
+#else
 static int bindfs_utimens(const char *path, const struct timespec ts[2])
+#endif
 {
     int res;
     char *real_path;
@@ -1095,7 +1293,7 @@ static int bindfs_utimens(const char *path, const struct timespec ts[2])
 #else
 #error "No symlink-compatible utime* function available."
 #endif
- 
+
     free(real_path);
     if (res == -1)
         return -errno;
@@ -1229,7 +1427,7 @@ static int bindfs_write(const char *path, const char *buf, size_t size,
 #ifdef __linux__
     if (source_buf != buf) {
         munmap(source_buf, mmap_size);
-    } 
+    }
 #endif
 
     return res;
@@ -1256,9 +1454,15 @@ static int bindfs_flock(const char *path, struct fuse_file_info *fi, int op)
     return 0;
 }
 
+#ifdef HAVE_FUSE_3
 static int bindfs_ioctl(const char *path, int cmd, void *arg,
                         struct fuse_file_info *fi, unsigned int flags,
                         void *data)
+#else
+static int bindfs_ioctl(const char *path, int cmd, void *arg,
+                        struct fuse_file_info *fi, unsigned int flags,
+                        void *data)
+#endif
 {
     int res = ioctl(fi->fh, cmd, data);
     if (res == -1) {
@@ -1490,7 +1694,9 @@ static struct fuse_operations bindfs_oper = {
     .init       = bindfs_init,
     .destroy    = bindfs_destroy,
     .getattr    = bindfs_getattr,
+    #ifndef HAVE_FUSE_3
     .fgetattr   = bindfs_fgetattr,
+    #endif
     /* no access() since we always use -o default_permissions */
     .readlink   = bindfs_readlink,
     .readdir    = bindfs_readdir,
@@ -1504,13 +1710,15 @@ static struct fuse_operations bindfs_oper = {
     .chmod      = bindfs_chmod,
     .chown      = bindfs_chown,
     .truncate   = bindfs_truncate,
+    #ifndef HAVE_FUSE_3
     .ftruncate  = bindfs_ftruncate,
+    #endif
     .utimens    = bindfs_utimens,
     .create     = bindfs_create,
     .open       = bindfs_open,
     .read       = bindfs_read,
     .write      = bindfs_write,
-#ifdef HAVE_FUSE_29
+#if defined(HAVE_FUSE_29) || defined(HAVE_FUSE_3)
     .lock       = bindfs_lock,
     .flock      = bindfs_flock,
 #endif
@@ -1865,8 +2073,8 @@ static int parse_map_file(UserMap *map, UserMap *reverse_map, char *file, int as
     UsermapStatus status;
 
     // Toggle between UID (passwd) and GID (group)
-    const int (*value_to_id)(const char *username, uid_t *ret);
-    const UsermapStatus (*usermap_add)(UserMap *map, uid_t from, uid_t to);
+    int (*value_to_id)(const char *username, uid_t *ret);
+    UsermapStatus (*usermap_add)(UserMap *map, uid_t from, uid_t to);
     const char *label_name, *label_id;
     if (as_gid) {
         value_to_id = &group_gid;
@@ -2114,6 +2322,48 @@ static void atexit_func()
     settings.mirrored_members = NULL;
 }
 
+struct fuse_args filter_special_opts(struct fuse_args *args)
+{
+    struct arena arena;
+    arena_init(&arena);
+    int tmp_argc = args->argc;
+    char **tmp_argv;
+
+    filter_o_opts(&keep_option, args->argc, (const char * const *)args->argv, &tmp_argc, &tmp_argv, &arena);
+
+    struct fuse_args new_args = FUSE_ARGS_INIT(0, (void*)0);
+    for (int i = 0; i < tmp_argc; i++) {
+        fuse_opt_add_arg(&new_args, tmp_argv[i]);
+    }
+
+    arena_free(&arena);
+    fuse_opt_free_args(args);
+    return new_args;
+}
+
+static bool keep_option(const char* opt)
+{
+    // Copied from "libfuse/util/mount.fuse.c" (fuse-3.10.1)
+    static const char * const ignored_opts[] = {
+        "",
+        "user",
+        "nofail",
+        "nouser",
+        "users",
+        "auto",
+        "noauto",
+        "_netdev",
+        NULL
+    };
+
+    for (int i = 0; ignored_opts[i] != NULL; ++i) {
+        if (strcmp(ignored_opts[i], opt) == 0) {
+            return false;
+        }
+    }
+    return true;
+}
+
 int main(int argc, char *argv[])
 {
     struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
@@ -2142,11 +2392,14 @@ int main(int argc, char *argv[])
         char *forward_odirect;
         char *uid_offset;
         char *gid_offset;
+        char *fsname;
     } od;
 
     #define OPT2(one, two, key) \
             FUSE_OPT_KEY(one, key), \
             FUSE_OPT_KEY(two, key)
+    #define OPT_OFFSET(opt, offset, key) \
+            {opt, offsetof(struct OptionData, offset), key}
     #define OPT_OFFSET2(one, two, offset, key) \
             {one, offsetof(struct OptionData, offset), key}, \
             {two, offsetof(struct OptionData, offset), key}
@@ -2215,8 +2468,9 @@ int main(int argc, char *argv[])
         OPT2("--enable-ioctl", "enable-ioctl", OPTKEY_ENABLE_IOCTL),
         OPT_OFFSET2("--multithreaded", "multithreaded", multithreaded, -1),
         OPT_OFFSET2("--forward-odirect=%s", "forward-odirect=%s", forward_odirect, -1),
-        OPT_OFFSET2("--uid-offset=%s", "uid-offset=%s", uid_offset, 0),
-        OPT_OFFSET2("--gid-offset=%s", "gid-offset=%s", gid_offset, 0),
+        OPT_OFFSET2("--uid-offset=%s", "uid-offset=%s", uid_offset, -1),
+        OPT_OFFSET2("--gid-offset=%s", "gid-offset=%s", gid_offset, -1),
+        OPT_OFFSET("fsname=%s", fsname, -1),
 
         FUSE_OPT_END
     };
@@ -2510,9 +2764,11 @@ int main(int argc, char *argv[])
     /* We want the kernel to do our access checks for us based on what getattr gives it. */
     fuse_opt_add_arg(&args, "-odefault_permissions");
 
+    // With FUSE 3 we set this in bindfs_init
+#ifndef HAVE_FUSE_3
     /* We want to mirror inodes. */
     fuse_opt_add_arg(&args, "-ouse_ino");
-    fuse_opt_add_arg(&args, "-oreaddir_ino");
+#endif
 
     /* Show the source dir in the first field on /etc/mtab, to be consistent
        with "real" filesystems.
@@ -2522,17 +2778,24 @@ int main(int argc, char *argv[])
        (issue #47). The character blacklist is likely not complete, which is
        acceptable since this is not a security check. The aim is to avoid giving
        the user a confusing error. */
-    if (strpbrk(settings.mntsrc, ", \t\n") == NULL) {
+    if (od.fsname != NULL) {
+        char *tmp = sprintf_new("-ofsname=%s", od.fsname);
+        fuse_opt_add_arg(&args, tmp);
+        free(tmp);
+    } else if (strpbrk(settings.mntsrc, ", \t\n") == NULL) {
         char *tmp = sprintf_new("-ofsname=%s", settings.mntsrc);
         fuse_opt_add_arg(&args, tmp);
         free(tmp);
     }
 
+    // With FUSE 3, we disable caches in bindfs_init
+#ifndef HAVE_FUSE_3
     /* We need to disable the attribute cache whenever two users
        can see different attributes. For now, only mirroring can do that. */
     if (is_mirroring_enabled()) {
         fuse_opt_add_arg(&args, "-oattr_timeout=0");
     }
+#endif
 
     /* If the mount source and destination directories are the same
        then don't require that the directory be empty.
@@ -2558,7 +2821,7 @@ int main(int argc, char *argv[])
         bindfs_oper.removexattr = NULL;
     }
 
-#ifdef HAVE_FUSE_29
+#if defined(HAVE_FUSE_29) || defined(HAVE_FUSE_3)
     /* Check that lock forwarding is not enabled in single-threaded mode. */
     if (settings.enable_lock_forwarding && !od.multithreaded) {
         fprintf(stderr, "To use --enable-lock-forwarding, you must use "
@@ -2585,6 +2848,9 @@ int main(int argc, char *argv[])
         bindfs_oper.ioctl = NULL;
     }
 
+    /* Remove/Ignore some special -o options */
+    args = filter_special_opts(&args);
+
     /* fuse_main will daemonize by fork()'ing. The signal handler will persist. */
     setup_signal_handling();
 
@@ -2593,5 +2859,5 @@ int main(int argc, char *argv[])
     fuse_opt_free_args(&args);
     close(settings.mntsrc_fd);
 
-    return fuse_main_return;
+    return bindfs_init_failed ? 1 : fuse_main_return;
 }
