@@ -18,6 +18,8 @@
 #   along with bindfs.  If not, see <http://www.gnu.org/licenses/>.
 #
 
+Dir.chdir(File.dirname(__FILE__))
+
 # if we are being run by make check it will set srcdir and we should use it
 $LOAD_PATH << (ENV['srcdir'] || '.')
 
@@ -27,7 +29,16 @@ require 'tempfile'
 
 include Errno
 
-$have_fuse_29 = Proc.new do
+$have_fuse_3 = Proc.new do
+  system("pkg-config --exists fuse3")
+  $?.success?
+end.call
+$have_fuse_3_readdir_bug = $have_fuse_3 && Proc.new do
+  system("pkg-config --max-version=3.10.1 fuse3")
+  $?.success?
+end.call
+
+$have_fuse_29 = !$have_fuse_3 && Proc.new do
   v = `pkg-config --modversion fuse`.split('.')
   raise "failed to get FUSE version with pkg-config" if v.size < 2
   v = v.map(&:to_i)
@@ -289,7 +300,9 @@ testenv("--chmod-deny --chmod-allow-x") do
 
     assert_exception(EPERM) { chmod(0777, 'mnt/file') }
     assert_exception(EPERM) { chmod(0000, 'mnt/file') }
-    assert_exception(EPERM) { chmod(01700, 'mnt/file') } # sticky bit
+    if `uname`.strip != 'FreeBSD'  # FreeBSD doesn't let us set the sticky bit on files
+      assert_exception(EPERM) { chmod(01700, 'mnt/file') } # sticky bit
+    end
 
     chmod(0611, 'mnt/file') # chmod that only changes x-bits should work
     assert { File.stat('src/file').mode & 07777 == 00611 }
@@ -396,8 +409,8 @@ root_testenv("--map=0/1:@0/@1", :title => "--map and chown/chgrp") do
     assert { File.stat('mnt/file1').gid == 1 }
 end
 
-Tempfile.create('passwdfile') do |passwd_file|
-    Tempfile.create('groupfile') do |group_file|
+Tempfile.open('passwdfile') do |passwd_file|
+    Tempfile.open('groupfile') do |group_file|
         passwd_file.puts("nobody:x:123:456:,,,:/tmp:/bin/false")
         passwd_file.flush
         group_file.write("#{nobody_group}:x:789")
@@ -457,18 +470,20 @@ testenv("", :title => "preserves inode numbers") do
     assert { File.stat('mnt/dir').ino == File.stat('src/dir').ino }
 end
 
-testenv("", :title => "preserves readdir inode numbers") do
-    touch('src/file')
-    mkdir('src/dir')
+unless $have_fuse_3_readdir_bug  # https://github.com/libfuse/libfuse/issues/583
+    testenv("", :title => "preserves readdir inode numbers") do
+        touch('src/file')
+        mkdir('src/dir')
 
-    inodes = {}
-    for line in `#{$tests_dir}/readdir_inode mnt`.split("\n").reject(&:empty?)
-        inode, name = line.split(" ")
-        inodes[name] = inode.to_i
+        inodes = {}
+        for line in `#{$tests_dir}/readdir_inode mnt`.split("\n").reject(&:empty?)
+            inode, name = line.split(" ")
+            inodes[name] = inode.to_i
+        end
+
+        assert { inodes['file'] == File.stat('src/file').ino }
+        assert { inodes['dir'] == File.stat('src/dir').ino }
     end
-
-    assert { inodes['file'] == File.stat('src/file').ino }
-    assert { inodes['dir'] == File.stat('src/dir').ino }
 end
 
 root_testenv("", :title => "setgid directories") do
@@ -701,7 +716,7 @@ testenv("", :title => "many files in a directory") do
   assert { Dir.entries('mnt/dir').sort == expected_entries.sort }
 end
 
-if $have_fuse_29
+if $have_fuse_29 || $have_fuse_3
   testenv("--enable-lock-forwarding --multithreaded", :title => "lock forwarding") do
     File.write('src/file', 'some contents for fcntl lockng')
     # (this test passes with an empty file as well, but this way is clearer)
@@ -751,7 +766,7 @@ end # have_fuse_29
 #
 # This test is also disabled for old (FUSE < 2.9) systems.
 # TODO: figure out why it doesn't work.
-if $have_fuse_29
+if $have_fuse_29 || $have_fuse_3
   root_testenv("--enable-ioctl", :title => "append-only ioctl", :valgrind => false) do
     touch('mnt/file')
     system('chattr +a mnt/file')
@@ -867,6 +882,13 @@ if `uname`.strip == 'Linux'
   end
 end
 
+# Issue 94
+if `uname`.strip != 'FreeBSD'  # -o fsname is not supported on FreeBSD
+  testenv("-o fsname=_bindfs_test_123_", :title => "fsname") do
+    assert { `mount` =~ /^_bindfs_test_123_\s+on\s+/m }
+  end
+end
+
 if `uname`.strip != 'FreeBSD'  # -o dev is not supported on FreeBSD
   root_testenv("-odev") do
     system("mknod mnt/zero c 1 5")
@@ -878,6 +900,14 @@ if `uname`.strip != 'FreeBSD'  # -o dev is not supported on FreeBSD
   end
 end
 
+# PR #95
+testenv("-ouser -onofail,nouser,,,delete-deny -o users -o auto,rename-deny,noauto -o chmod-deny,_netdev,,", :title => "filtering special options") do
+  touch('src/file')
+  assert_exception(EPERM) { rm('mnt/file') }
+  assert_exception(EPERM) { File.rename('mnt/file', 'mnt/file2') }
+  assert_exception(EPERM) { chmod(0777, 'mnt/file') }
+end
+
 # FIXME: this stuff around testenv is a hax, and testenv may also exit(), which defeats the 'ensure' below.
 # the test setup ought to be refactored. It might well use MiniTest or something.
 # TODO: support FreeBSD in this test (different group management commands)
